//Ethan Alexander Shulman 2022 "Void Consumes Moon" April 3, 2022

//scene rendering options
#define VIEW_STEPS 30
#define VIEW_STEP_DITHER .1
#define VIEW_EPSILON 1e-2
//#define VIEW_EPSILON_MUL 1.14
#define LIGHTING_STEPS 30
#define LIGHTING_STEP_DITHER .1
#define LIGHTING_EPSILON 1e-2
//#define COLOR_MAX 100.
#define DENOISE_RANGE 5
#define DENOISE_NORMAL_THRESHOLD .99
#define DENOISE_DEPTH_THRESHOLD 1e1
#define LOW_DISCREPANCY 1
//#define ANIMATION_TIME 30.
#define ANIMATION_SAMPLES 100
#ifdef ANIMATION_TIME
out float time;
#else
float time = 0.;
#endif

//constants
#define PI 3.14159265





mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))



/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	return mat3(0,0,0,
			0,0,(cos(time/30.*PI)*.5+.5)*PI*2.,
			3.5,250,-400);
}


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*dot(abs(posSeed),baseRand.zxy)*baseRand.xy)
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = mod(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy),3.)
#define CHROMA(mt,rc,gc,bc) if (chromaFrame < 1.) {mt[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mt[0]*=vec3(0,3,0);gc;}\
else {mt[0]*=vec3(0,0,3);bc;}



//functions defining scene

/*javascript to randomly generate nice procedural cosine noise

//CONFIG
var MAX_DEPTH = 2,
DEPTH_CHANCE = 0.99,
MAX_FREQ = 20,
MIN_FREQ = 0.1,
FREQ_EXP = 1,
DIMENSIONS = 3,
POS_VARNAME = "p";

//generation code
var txt = "";
function gen(depth) {
	txt += "cos(dot("+POS_VARNAME+",vec"+DIMENSIONS+"(";
	for (var i = 0; i < DIMENSIONS; i++) {
		if (i) txt += ",";
		txt += (Math.pow(Math.random(),FREQ_EXP)*(MAX_FREQ-MIN_FREQ)+MIN_FREQ)*Math.PI*2/DIMENSIONS;
	}
	txt += "))";
	if (depth < MAX_DEPTH && Math.random() < DEPTH_CHANCE) {
		txt += "+";
		depth++;
		gen(depth);
		txt += "*"+(Math.pow(Math.random(),FREQ_EXP)*(MAX_FREQ-MIN_FREQ)+MIN_FREQ)*Math.PI*2;
	}
	txt += ")";
}
gen(0);
console.log(txt);
*/

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	vec2 geo = vec2(length(p)-100., 0.);//moon
	if (geo.x < 1.) {
		for (float f = 1.; f < 40.; f += f*1.3333) {
			vec3 fp = floor(10000.+(p/f));
			geo.x = max(geo.x, -(length(mod(abs(p),f)-f*.5)-f*fract(dot(fp,vec3(.2874,.37642,.3645))+dot(fp,vec3(.0432874,.0187234,.037676234)))*.5));
		}
	}
	return geo;
}

float voronoi(vec3 u, float i) {
	#define l(i) length(fract(abs(u)*.01+fract(i*vec3(1,8,5))+cos(u.yzx*.03+i*8.))-.5)
	return min(l(i),min(l(i+1.),l(i+2.)));
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;
float refractionIndex;*/
mat3 material(vec3 p, vec3 n, float mat) {
	//moon
	float v = 1e8, cl = length(p-normalize(vec3(.3,.5,-1))*100.);
	vec3 vp = p+100.;
	for (float f = 1.; f < 10.; f += f*.444) {
		v = min(v,voronoi(vp*f,0.));//void pattern
	}
	v = clamp(v*2.-.5+(cl+10.-pow(time/30.,8.)*120.)*.01,0.,1.);//void animation
	return mat3(vec3(.65,.65,.7)*v, .24,0,1, 1,0,0);
}

//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	//stars
	vec4 hv = hash(uvec4(floor(d/.006+10000.),0))*I2F;
	vec3 sv = max(hv.w-.95,0.)*20.*(.5+.5*hv.xyz);
	return clamp(1.-(length(mod(abs(d),.006)-.003)-.0001)*2000., 0.,1.)*sv+
		vec3(1,.98,.95)*clamp(1.-(length(d-normalize(vec3(1,1,-1)))-.1)*100.,0.,1.)*100.;//sun
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return background(p,d)*clamp(length(p)-100.,0.,1.);
}

